/*----------------------------------------------------------------------
    Document Image Labeling                   dlabel  ver1.21
        Written by  H.Goto, Mar. 1999
        Modified by H.Goto, June 1999
        Modified by H.Goto, Sep. 1999
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1999  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


//#define		FastXlatMode
#define		Merge8NNMode
#define		Merge5x5Mode

#define		SUBREGIONSIZE	50
#define		WINDOWSIZE	60
#define		WINDOWOFFSET	-5
#define		MaxClusters	5

#define		TruncateTinyGroups
#define		TruncLevel	16
#define		TruncCounts	4

#define		IMGBUFDEPTH	8

#define		GrayLevels	256
#define		MAXPEAKS	(GrayLevels /4)

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <math.h>

#include	"utypes.h"

#include	"CRect.h"
#include	"objgrp.h"
#include	"ufilep.h"
#include	"siplib.h"
#include	"xiplib.h"
#include	"sgplib.h"
#include	"iset.h"
#include	"labelset.h"
#include	"filters.h"
#include	"gettime.h"
#include	"findcol_com.h"
#include	"findgray_pd.h"


		/* minimum difference at peak merging */
static int	mindiff[] = { 10,20,30,-1 };

static int	sw_verbose = 0;
static int	sw_weighted = 0;
static int	sw_smooth = 0;

static double	utime_load,stime_load;
static double	utime_filt,stime_filt;
static double	utime_edge,stime_edge;
static double	utime_hist,stime_hist;
static double	utime_merge,stime_merge;
static double	utime_split,stime_split;


typedef unsigned short	GLAB;

typedef struct {
	int		ncolor;
	CCluster	collist[MAXPEAKS];
	int		glabel[MAXPEAKS];
} MTCELL;




/*----------------------------------------------
    Find local gray levels
----------------------------------------------*/

static void find_graylevels(SIPImage *image,MTCELL **cmatrix, \
		int mwidth,int mheight,int weighted){
	int	mx,my;
	int	i,c,l;
	int	n_cluster;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			n_cluster = findgray_pd(image, \
				SUBREGIONSIZE * mx + WINDOWOFFSET, \
				SUBREGIONSIZE * my + WINDOWOFFSET, \
				WINDOWSIZE,WINDOWSIZE, \
				cmatrix[my][mx].collist,MAXPEAKS,weighted);
			if ( n_cluster < 0 )  n_cluster = 0;
			cmatrix[my][mx].ncolor = n_cluster;
#if 0
			printf("cell(%d,%d): ncolor=%d ",mx,my,cmatrix[my][mx].ncolor);
			for ( i=0 ; i<cmatrix[my][mx].ncolor ; i++ ){
				printf(" %d",cmatrix[my][mx].collist[i].g);
			}
			printf("\n");
#endif
		}
	}
	return;
}




/*----------------------------------------------
    Make gray-to-label translation table
----------------------------------------------*/

static int make_xlat(MTCELL *cmatp,GLAB *xlat_lab,int *glrev){
	int	i,j;
	int	divpos[MAXPEAKS +1];
	divpos[0] = 0;
	for ( i=0 ; i < cmatp->ncolor -1 ; i++ ){
		divpos[i+1] = (cmatp->collist[i+1].g \
				+ cmatp->collist[i].g) /2;
	}
	divpos[cmatp->ncolor] = GrayLevels;

	for ( j=0 ; j<cmatp->ncolor ; j++ ){
		glrev[ cmatp->glabel[j] ] = j;
		for ( i=divpos[j] ; i<divpos[j+1] ; i++ ){
			xlat_lab[i] = cmatp->glabel[j];
		}
	}
	return(0);
}




/*----------------------------------------------
    Find close gray id.
----------------------------------------------*/

static int find_closegray(MTCELL *cmatp,int gray){
	int	i;
	int	mindist,minpos,dist;
	mindist = 0x7fffffff;  minpos = -1;
	for ( i=0 ; i < cmatp->ncolor ; i++ ){
		dist = abs(gray - cmatp->collist[i].g);
		if ( mindist > dist ){
			mindist = dist;  minpos = i;
		}
	}
	return(minpos);
}




/*----------------------------------------------
    Make reduced image (subroutine)
----------------------------------------------*/

static int make_reducedimage(SIPImage *labelimage,SIPImage *rimage, \
		MTCELL **cmatrix,int mwidth,int mheight,int *glrev){
	int	mx,my,x,y,i;
	GLAB	*src_lab;
	uchar	*dst;
	MTCELL	*cmatp;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];

			for ( y=0 ; y<SUBREGIONSIZE ; y++ ){
				src_lab = (GLAB *)sip_getimgptr(labelimage, \
					y + my * SUBREGIONSIZE);
				src_lab += mx * SUBREGIONSIZE;
				dst = (uchar *)sip_getimgptr(rimage, \
					y + my * SUBREGIONSIZE);
				dst += mx * SUBREGIONSIZE;
				for ( x=0 ; x<SUBREGIONSIZE ; x++ ){
					dst[x] = (uchar)cmatp->collist[ glrev[src_lab[x]] ].g;
				}

			}
		}
	}
	return(0);
}




/*----------------------------------------------
    Make label image (& reduced image)
----------------------------------------------*/

static int make_labelimage(SIPImage *image,SIPImage *rimage,SIPImage *labelimage, \
		MTCELL **cmatrix,int mwidth,int mheight,int glcount){
	int	mx,my;
	int	x,y,i;
	uchar	*src,*dst;
	GLAB	*dst_lab;
	int	*glrev;
	int	minpos;
	int	xlat[GrayLevels];
	GLAB	xlat_lab[GrayLevels];
	MTCELL	*cmatp;
	if ( 0 == (glrev = new int[ glcount > 0 ? glcount : 1 ]) )  return(-1);
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];
#ifdef FastXlatMode
			make_xlat(cmatp,xlat_lab,glrev);
#endif
			for ( y=0 ; y<SUBREGIONSIZE ; y++ ){
				src = (uchar *)sip_getimgptr(image, \
					y + my * SUBREGIONSIZE);
				src += mx * SUBREGIONSIZE;
				dst_lab = (GLAB *)sip_getimgptr(labelimage, \
					y + my * SUBREGIONSIZE);
				dst_lab += mx * SUBREGIONSIZE;
				for ( x=0 ; x<SUBREGIONSIZE ; x++ ){
#ifndef FastXlatMode
					minpos = find_closegray(cmatp,\
							(int)((uint)src[x]));
					dst_lab[x] = cmatp->glabel[minpos];
					glrev[ dst_lab[x] ] = minpos;
#else
					dst_lab[x] = xlat_lab[ (int)((uint)src[x]) ];
#endif	/* FastXlatMode */
				}
			}
		}
	}

	if ( rimage ){
		make_reducedimage(labelimage,rimage,\
			cmatrix,mwidth,mheight,glrev);
	}

	delete []glrev;
	return(0);
}




/*----------------------------------------------
    Check conflict of global labels
----------------------------------------------*/

static int isconflict(int *glmap,LSETS *lsets,int gl1,int gl2,int maxcount){
	int	sid1,sid2,i,j;
	int	count;
	IDXSET	*idxset1,*idxset2;
	sid1 = lsets->revtbl[gl1];
	sid2 = lsets->revtbl[gl2];
	idxset1 = lsets->settbl[sid1];
	idxset2 = lsets->settbl[sid2];
	for ( i=count=0 ; i<idxset1->counts ; i++ ){
		for ( j=0 ; j<idxset2->counts ; j++ ){
			if ( glmap[ idxset1->idlist[i] ] \
			  == glmap[ idxset2->idlist[j] ] )  count++;
			if ( count >= maxcount )  break;
		}
		if ( count >= maxcount )  break;
	}
	return(count);
}




/*----------------------------------------------
    Merge labels
----------------------------------------------*/

static int mergelabel_2(LSETS *lsets,int *glmap, \
		MTCELL *cmatp1,MTCELL *cmatp2,int diff,int conflict){
	int	i,j;
	int	gl1,gl2;
	int	ov;
	int	dr,dg,db;
	for ( i=0 ; i < cmatp1->ncolor ; i++ ){
		for ( j=0 ; j < cmatp2->ncolor ; j++ ){
			if ( abs(cmatp1->collist[i].g - cmatp2->collist[j].g) \
				>= diff )  continue;

			gl1 = cmatp1->glabel[i];  gl2 = cmatp2->glabel[j];

			if ( conflict != -1 ){
				if ( isconflict(glmap,lsets,gl1,gl2,\
					conflict +1) > conflict )  continue;
			}

			if ( gl1 > gl2 ){
				gl2 = cmatp1->glabel[i];  gl1 = cmatp2->glabel[j];
			}
			if ( 0 > lsets->mergelabel(gl1,gl2) ){
				fprintf(stderr,"mergelabel() failed.\n");  return(-1);
			}
		}
	}
	return(0);
}


static int mergelabel(LSETS *lsets,int glcount,int *glmap, \
		MTCELL **cmatrix,int mwidth,int mheight){
	int	i,id,sid,count;
	int	mx,my;
	IDXSET	*idxset;
	int	glevel;
	int	single_label[ (GrayLevels/TruncLevel) +1];

	/* At the begining, each global label should belong to a label set. */
	for ( i=0 ; i<glcount ; i++ ){
		if ( 0 > (id = lsets->newset() ) ){
			fprintf(stderr,"newset() failed.\n");  return(-1);
		}
		if ( 0 > lsets->addmember(id,i) ){
			fprintf(stderr,"addmember() failed.\n");  return(-1);
		}
	}

	/* merge labels among adjacent (upper,left) cells */
	for ( i=0 ; mindiff[i]>=0 ; i++ ){
		for ( my=0 ; my<mheight ; my++ ){
			for ( mx=0 ; mx<mwidth ; mx++ ){
				if ( mx > 0 ){
					if ( mergelabel_2(lsets,glmap,&cmatrix[my][mx], \
						&cmatrix[my][mx-1],mindiff[i],2) )  return(-1);
				}
				if ( my > 0 ){
					if ( mergelabel_2(lsets,glmap,&cmatrix[my][mx], \
						&cmatrix[my-1][mx],mindiff[i],2) )  return(-1);
				}
#ifdef Merge8NNMode
				if ( mx > 0 && my > 0 ){
					if ( mergelabel_2(lsets,glmap,&cmatrix[my][mx], \
						&cmatrix[my-1][mx-1],mindiff[i],2) )  return(-1);
				}
				if ( mx < mwidth -1 && my > 0 ){
					if ( mergelabel_2(lsets,glmap,&cmatrix[my][mx], \
						&cmatrix[my-1][mx+1],mindiff[i],2) )  return(-1);
				}
#endif
			}
		}
	}

#ifdef	Merge5x5Mode
	/* merge labels among adjacent (+2 upper, +2 left) cells */
	for ( i=0 ; mindiff[i]>=0 ; i++ ){
		for ( my=0 ; my<mheight ; my++ ){
			for ( mx=0 ; mx<mwidth ; mx++ ){
				if ( mx > 1 ){
					if ( mergelabel_2(lsets,glmap,&cmatrix[my][mx], \
						&cmatrix[my][mx-2],mindiff[i],2) )  return(-1);
				}
				if ( my > 1 ){
					if ( mergelabel_2(lsets,glmap,&cmatrix[my][mx], \
						&cmatrix[my-2][mx],mindiff[i],2) )  return(-1);
				}
			}
		}
	}
#endif

#ifdef	TruncateTinyGroups
	/* truncate tiny groups */
	for ( i=0 ; i <= (GrayLevels/TruncLevel) ; i++ )  single_label[i] = -1;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			for ( i=0 ; i<cmatrix[my][mx].ncolor ; i++ ){
				sid = lsets->revtbl[ cmatrix[my][mx].glabel[i] ];
				idxset = lsets->settbl[sid];
				if ( idxset->counts >= TruncCounts )  continue;
				glevel = cmatrix[my][mx].collist[i].g / TruncLevel;
				if ( single_label[glevel] != -1 ){
 					lsets->mergelabel(single_label[glevel], \
						idxset->idlist[0]);
				}
				else{	single_label[glevel] = idxset->idlist[0];
				}
			}
		}
	}
#endif

	if ( sw_verbose ){
		for ( sid=count=0 ; sid<glcount ; sid++ ){
			if ( 0 != lsets->settbl[sid] )  count++;
		}
		printf("Number of labels after merging: %d\n",count);
	}
	return(0);
}




/*----------------------------------------------
    Assign global labels
----------------------------------------------*/

static int assign_labels(MTCELL **cmatrix,int mwidth,int mheight){
	int	mx,my,j;
	MTCELL	*cmatp;
	GLAB	glcount = 0;	/* global label counter */
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];
			for ( j=0 ; j<cmatp->ncolor ; j++ ){
				cmatp->glabel[j] = glcount++;
			}
		}
	}
	return(glcount);
}




/*----------------------------------------------
    Make global label mapping table
        (global label -> subregion ID.)
----------------------------------------------*/

static int * mk_glmap(int glcount,MTCELL **cmatrix,int mwidth,int mheight){
	int	mx,my,i;
	MTCELL	*cmatp;
	int	*glmap;
	if ( 0 == (glmap = new int[glcount]) )  return(0);
	for ( i=0 ; i<glcount ; i++ )  glmap[i] = -1;
	for ( my=0 ; my<mheight ; my++ ){
		for ( mx=0 ; mx<mwidth ; mx++ ){
			cmatp = &cmatrix[my][mx];
			for ( i=0 ; i<cmatp->ncolor ; i++ ){
				glmap[ cmatp->glabel[i] ] = my * mwidth + mx;
			}
		}
	}
	return(glmap);
}




/*----------------------------------------------
    Make split images
----------------------------------------------*/

static int make_splitimages(SIPImage *labelimage,LSETS *lsets, \
		int glcount,int width,int height){
	SIPImage	*simage;
	GLAB	*src;
	uchar	*dst;
	int	x,y;
	int	sid,count,retcode;
	char	outfile[30];
	IDXSET	*idxset;
	PBMFILE	PBMSV;
	if ( 0 == (simage = sip_CreateImage(width,height,8)) )  return(-1);
	
	for ( sid=count=0 ; sid<glcount ; sid++ ){
		if ( 0 == (idxset = lsets->settbl[sid]) )  continue;
		sip_ClearImage(simage,1);

		for ( y=0 ; y<height ; y++ ){
			src = (GLAB *)sip_getimgptr(labelimage,y);
			dst = (uchar *)sip_getimgptr(simage,y);
			for ( x=0 ; x<width ; x++ ){
				if ( lsets->revtbl[ (int)src[x] ] == sid ){
					dst[x] = 0;
				}
			}
		}

		sprintf(outfile,"%04d.pbm",count);
		PBMSV.setsize(width,height,1);
		retcode = PBMSV.open(outfile,"w");
		if ( 0 > retcode ){
			sip_DestroyImage(simage);  return(-1);
		}
		retcode = 0;
		for ( y=0 ; y<height ; y++ ){
			if ( PBMSV.writeline_bilevel(-1, \
				(void *)sip_getimgptr(simage,y),1) ){
				retcode = -1;  break;
			}
		}
		if ( retcode ) PBMSV.close();  else retcode = PBMSV.close();
		if ( retcode ){
			sip_DestroyImage(simage);  return(-1);
		}

		count++;
	}

	sip_DestroyImage(simage);
	return(0);
}




/*----------------------------------------------
    Load PBM/PGM/PPM image file
----------------------------------------------*/

static int load_image(char *fname,SIPImage **image, \
		int *width,int *height,int bufdepth){
	PBMFILE	PBMLD;
	int	retcode,depth,x,y;
	SIPImage	*imgbuf;
	uchar	*lbuf,*src,*dst;
	uint	c;
	int	bufwidth,bufheight;
	*image = 0;	/* default value = NULL */
	PBMLD.setpal_fb(0,0xff);
	if ( bufdepth != 8 && bufdepth != 32 )  return(-5);
	retcode = PBMLD.open(fname,"r");
	if ( retcode )  return(retcode);
	switch ( PBMLD.getfiletype() ){
	  case 0:	depth = 1;  break;
	  case 1:	depth = 8;  break;
	  case 2:	depth = 32;  break;	/* 24bit */
	  default:	return(-5);
	}
	*width  = PBMLD.getwidth();
	*height = PBMLD.getheight();

	bufwidth = ((*width + SUBREGIONSIZE -1) / SUBREGIONSIZE) * SUBREGIONSIZE;
	bufheight = ((*height + SUBREGIONSIZE -1) / SUBREGIONSIZE) * SUBREGIONSIZE;

	if ( 0 == (lbuf = new uchar[3 * bufwidth]) )  return(-6);
	if ( 0 == (imgbuf = sip_CreateImage(bufwidth,bufheight,bufdepth)) ){
		delete []lbuf;
		return(-6);
	}
	sip_ClearImage(imgbuf,0xff);
	retcode = 0;
	for ( y=0 ; y < *height ; y++ ){
		if ( depth == 1 ){
			if ( 0 != PBMLD.readline_gray(-1,(void *)lbuf) ){
				retcode = -1;	break;	/* read error */
			}
		}
		else{	/* depth=8, 32 */
			if ( 0 != PBMLD.readline(-1,(void *)lbuf) ){
				retcode = -1;	break;	/* read error */
			}
		}
		if ( 32 == depth ){
			src = lbuf;  dst = (uchar *)sip_getimgptr(imgbuf,y);
			if ( bufdepth == 8 ){
				for ( x=0 ; x < *width ; x++ ){
					c = (uint)(.299 * (double)((uint)src[0]) \
						 + .587 * (double)((uint)src[1]) \
						 + .114 * (double)((uint)src[2]));
					if ( c > 255 )  c = 255;
					*dst++ = c;
					src += 3;
				}
			}
			else{	for ( x=0 ; x < *width ; x++ ){
					dst[0] = src[0];
					dst[1] = src[1];
					dst[2] = src[2];
					dst[3] = 0;
					src += 3;  dst += 4;
				}
			}
		}
		else{	if ( bufdepth == 8 ){
				memcpy((void *)sip_getimgptr(imgbuf,y), \
					(void *)lbuf,*width);
			}
			else{	src = lbuf;
				dst = (uchar *)sip_getimgptr(imgbuf,y);
				for ( x=0 ; x < *width ; x++ ){
					dst[0] = dst[1] = dst[2] = *src++;
					dst[3] = 0;
					dst += 4;
				}
			}
		}
	}
	delete []lbuf;
	PBMLD.close();
	if ( ! retcode ){
		*image = imgbuf;
	}
	else{	sip_DestroyImage(imgbuf);
	}
	return(retcode);
}




/*----------------------------------------------
    Print usage
----------------------------------------------*/

static void usage(){
	fputs("dlabel v1.21, Copyright (C) 1999 Hideaki Goto \n",stderr);
	fputs("Usage: dlabel [-options] in_file \n",stderr);
	fputs("         -s      :  smooth input image \n",stderr);
	fputs("         -w      :  use weighted-histogram \n",stderr);
	fputs("         -ri file:  save reduced image \n",stderr);
	fputs("         -verbose:  show processing information \n",stderr);
}




/*----------------------------------------------
    Main routine
----------------------------------------------*/

int main(int ac,char *av[]){
	char	*infile  = NULL;
	char	*rimagefile = 0;
	int	width,height,mwidth,mheight;
	int	i,j,k;
	SIPImage	*image;
	SIPImage	*rimage = 0;
	SIPImage	*labelimage;
	MTCELL	**cmatrix;
	int	glcount;
	int	*glmap;
	LSETS	LSETS;

	for ( k=1, i=1 ; i<ac && k ; i++ ){
		if ( 0 == strcmp("-ri",av[i]) ){
			if ( i++ >= ac ){  k=0; continue; }
			rimagefile = av[i];
			continue;
		}
		if ( 0 == strcmp("-s",av[i]) ){  sw_smooth = -1;  continue; }
		if ( 0 == strcmp("-w",av[i]) ){  sw_weighted = -1;  continue; }
		if ( 0 == strcmp("-verbose",av[i]) ){  sw_verbose = -1;  continue; }
		if ( 0 == strcmp("-v",av[i]) ){  sw_verbose = -1;  continue; }
		if ( av[i][0] == '-' ){  k=0;  continue; }
		if ( NULL == infile ){  infile = av[i];  continue; }
	}
	if ( k == 0 || infile == NULL ){  usage();  return(0); }


/* ---- Read Image ---- */

	time_start();
	if ( 0 > load_image(infile,&image,&width,&height,IMGBUFDEPTH) ){
		fprintf(stderr,"Can't read \"%s\".\n",infile);
		return(2);
	}
	time_get(&utime_load,&stime_load);


/* ---- Allocate Buffers ---- */

	mwidth  = image->width  / SUBREGIONSIZE;
	mheight = image->height / SUBREGIONSIZE;
	if ( sw_verbose ){
		printf("Image  size = %d x %d\n",width,height);
		printf("Buffer size = %d x %d\n",image->width,image->height);
		printf("Matrix size = %d x %d\n",mwidth,mheight);
	}

	if ( rimagefile ){
		if ( 0 == (rimage = sip_CreateImage(image->width,image->height,IMGBUFDEPTH)) ){
			fprintf(stderr,"Low memory.\n");
			return(1);
		}
	}

	if ( 0 == (labelimage = sip_CreateImage(image->width, \
				image->height,sizeof(GLAB) * 8)) ){
		fprintf(stderr,"Low memory.\n");
		return(1);
	}

	if ( 0 == (cmatrix = new MTCELL*[ mheight ]) ){
		fprintf(stderr,"Low memory.\n");
		return(1);
	}
	for ( i=0 ; i<mheight ; i++ ){
		if ( 0 == (cmatrix[i] = new MTCELL[ mwidth ]) ){
			fprintf(stderr,"Low memory.\n");
			return(1);
		}
	}


/* ---- Image Filtering ---- */

	utime_filt = stime_filt = 0;
	if ( sw_smooth ){
		time_start();
		smooth_image(image,image);
		if ( sw_verbose )  printf("Image filtering has done.\n");
		time_get(&utime_filt,&stime_filt);
	}


/* ---- Create Gray Level Histograms ---- */

	utime_edge = stime_edge = 0;
	if ( sw_weighted ){
		time_start();
		if ( 0 > findgray_pd_setweight(image) ){
			fprintf(stderr,"Low memory.\n");
			return(1);
		}
		if ( sw_verbose )  printf("Edge detection has done.\n");
		time_get(&utime_edge,&stime_edge);
	}

	time_start();
	find_graylevels(image,cmatrix,mwidth,mheight,sw_weighted);
	glcount = assign_labels(cmatrix,mwidth,mheight);
	if ( sw_verbose )  printf("Gray level detection has done.\n");
	time_get(&utime_hist,&stime_hist);


/* ---- Merge Labels ---- */

	time_start();
	if ( 0 == (glmap = mk_glmap(glcount,cmatrix,mwidth,mheight)) ){
		fprintf(stderr,"Low memory.\n");
		return(1);
	}
	mergelabel(&LSETS,glcount,glmap,cmatrix,mwidth,mheight);
	time_get(&utime_merge,&stime_merge);


/* ---- Make Label Image ---- */

	time_start();
	make_labelimage(image,rimage,labelimage,cmatrix,mwidth,mheight,glcount);
	if ( rimage ){
		xip_SaveImage(rimagefile,rimage);
	}


/* ---- Make Split Image ---- */

	if ( 0 > make_splitimages(labelimage,&LSETS,glcount,width,height) ){
		fputs("Can't write output file.\n",stderr);
		return(2);
	}
	time_get(&utime_split,&stime_split);

	if ( sw_verbose ){
		printf("Time: Image_loading	%.3f(usr) %.3f(sys)\n", \
						utime_load,stime_load);
		if ( sw_smooth )
		  printf("Time: Image_filtering	%.3f(usr) %.3f(sys)\n", \
						utime_filt,stime_filt);
		if ( sw_weighted )
		  printf("Time: Edge_detection	%.3f(usr) %.3f(sys)\n", \
						utime_edge,stime_edge);
		printf("Time: Gray_detection	%.3f(usr) %.3f(sys)\n", \
						utime_hist,stime_hist);
		printf("Time: Label_merge	%.3f(usr) %.3f(sys)\n", \
						utime_merge,stime_merge);
		printf("Time: Mk_split_image	%.3f(usr) %.3f(sys)\n", \
						utime_split,stime_split);
		printf("Time: Total		%.3f(usr) %.3f(sys)\n", \
			(float)(utime_load + utime_filt + utime_edge \
			+ utime_hist + utime_merge + utime_split), \
			(float)(stime_load + stime_filt + stime_edge \
			+ stime_hist + stime_merge + stime_split) );
	}

	sip_DestroyImage(labelimage);
	sip_DestroyImage(rimage);
	sip_DestroyImage(image);
	return(0);
}


