.TH dlabel 1 "Nov. 1999" "Version 1.21"
.SH NAME
dlabel \- document image labeling tool for character pattern extraction
(v1.21)
.SH SYNOPSIS
.B dlabel
.BR
[\-options]
.BR
.I in_file

.SH DESCRIPTION
dlabel is the tool for extracting character patterns from 
grayscale document images with complex background.
dlabel can separate character patterns of various gray levels and sizes
from overlapping background images as far as there is 
enough difference of gray values between the character patterns
and the background images.

.PP
dlabel reads a document image of raw, pgm(5) format from
.I in_file,
segments locally\-isochromatic objects such as character patterns,
groups the objects,
and writes each object groups to individual raw, pbm(5) files.
The name of the output files are 0000.pbm, 0001.pbm, ..., and 
they are put into the current directory.
dlabel can also read a document image of raw, ppm(5) format,
however, the image is converted to the grayscale image internally.


.PP
The simplified descriptions of the processes in this program
are as shown below.
(See references for more detail of the processes.)

.TP
1.
The input image is partitioned into small (50x50dots) subregions.
.TP
2.
Local representative gray levels are found in every subregion
analyzing the gray level histogram.
The input image is segmented by gray level using multilevel thresholding.
.TP
3.
The locally segmented objects are merged between neighboring 
subregions according to the difference of the corresponding 
representative gray levels.
.TP
4.
Each group of merged objects is written to an bitmap file,
regarding the objects are black (foreground) images.

.PP
The method used in this program has following advantageous
properties.

.TP
1.
It can separate light character patterns and dark ones
simultaneously.
.TP
2.
It can extract very thin (>1.5pixels) character strokes.
.TP
3.
It is tolerant of dull image in which edges are not 
very clear.
.TP
4.
It is tolerant of slight shading of image.
Even if the brightness of a text line changes gradually,
it can extract the pattern of the text line as a single image.
(The whole pattern of the text line will appear in an output file.)


.SH OPTIONS
.TP
.B \-s
Apply a simple smoothing filter to the input image.
This option is useful for some noisy images.
.TP
.B \-w
Use weighted histogram when dlabel finds local representative gray levels.
This option is very useful for extraction of thin character strokes.
Users are strongly recommended to specify this option.
.TP
.B \-ri file
Let dlabel write locally reduced image to 
.I file
in pgm(5) format.
.TP
.B \-verbose, \-v
Let dlabel show further processing information during its process.


.SH EXIT STATUS
The following exit values are returned:
.PD 0
.TP 10
.B 0
Successful completion.
.TP
.B > 0
An error occurred.
.PD

.SH SEE ALSO
pbm(5), pgm(5), ppm(5)

.SH NOTES
This program can separate character patterns from background images,
however, it cannot distinguish which images are characters and which are
background.

.PP
This program is not very tolerant of the moire noise caused by
interference between halftone image and image sampling grid.
This program does not work well for very noisy images.
An appropriate noise filter is sometimes required as a preprocessing.

.PP
Since this program was designed specially for character pattern extraction
from document images, it is not suitable for segmentation of general images.

.PP
This program was designed so that it would not oversegment character 
patterns or text line patterns as far as possible.
However, background images are often oversegmented greatly.
The program would produce a lot of output files which consist of
oversegmented pieces of image.

.PP
Unfortunately,
this program may not work well for some images in which
objects (including characters) of various gray levels apper densely; 
such as bookshelf images taken by a camera.


.SH REFERENCES

Hideaki Goto, Masatsugu Hirayama and Hirotomo Aso, 
"Character pattern extraction from grayscale document images
using local multilevel thresholding,"
Trans. IEICE (D\-II), Vol.J82\-D\-II, No.11, pp.2188\-2192, 1999.
(in Japanese)

.SH AUTHOR
Copyright (C) 1999 Hideaki Goto
.\" Permission to use, copy, modify, and distribute this software and
.\" its documentation for any purpose is hereby granted without fee,
.\" provided that (i) the above copyright notice and this permission
.\" notice appear in all copies and in supporting documentation, (ii)
.\" the name of the author, Hideaki Goto, may not be used in any
.\" advertising or otherwise to promote the sale, use or other
.\" dealings in this software without prior written authorization
.\" from the author, (iii) this software may not be used for
.\" commercial products without prior written permission from the
.\" author, and (iv) the notice of the modification is specified in
.\" case of that the modified copies of this software are distributed.
.\"
.\" THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
.\" THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
.\" SOFTWARE.

E\-Mail: hgot@aso.ecei.tohoku.ac.jp, hgot@ecip.tohoku.ac.jp
